@@ -16,11 +16,13 @@
 #include <dirent.h>
 #include <execinfo.h>
 #include <ctype.h>
+#include <sys/wait.h>
+#include <endian.h>
 #include "dr.h"
 #include "ofdt.h"
 
-char *dlpar_add_slot = DLPAR_ADD_SLOT;
-char *dlpar_remove_slot = DLPAR_REMOVE_SLOT;
+char *add_slot_fname = ADD_SLOT_FNAME;
+char *remove_slot_fname = REMOVE_SLOT_FNAME;
 
 #define DR_LOG_PATH	"/var/log/drmgr"
 #define DR_LOG_PATH0	"/var/log/drmgr.0"
@@ -28,38 +30,91 @@ char *dlpar_remove_slot = DLPAR_REMOVE_SLOT;
 #define LPARCFG_PATH	"/proc/ppc64/lparcfg"
 
 static int dr_lock_fd = 0;
+static long dr_timeout;
 
 /**
- * set_debug
- * @brief Common routine to set the debug level
+ * set_output level
+ * @brief Common routine to set the output level
  *
- * @param dbg_level debug level to set debug to
+ * @param level level to set the output level to
  */
 inline void
-set_debug(int dbg_level)
+set_output_level(int level)
 {
-	debug = dbg_level;
+	output_level = level;
 
-	if (dbg_level >= 14) {
-		dbg("Enabling RTAS debug\n");
-		rtas_set_debug(dbg_level);
+	if (output_level >= 14) {
+		say(DEBUG, "Enabling RTAS debug\n");
+		rtas_set_debug(output_level);
 	}
 }
 
+int say(enum say_level lvl, char *fmt, ...)
+{
+	va_list ap;
+	char buf[256];
+	int len, rc;
+
+	va_start(ap, fmt);
+	memset(buf, 0, 256);
+	len = vsnprintf(buf, 256, fmt, ap);
+	va_end(ap);
+
+	if (len >= 256) {
+		strcpy(&buf[243], "<truncated>\n");
+		len = 255;
+	}
+
+	if (log_fd)
+		rc = write(log_fd, buf, len);
+
+	if (lvl <= output_level)
+		fprintf(stderr, "%s", buf);
+
+	return len;
+}
+
+void report_unknown_error(char *file, int line) {
+	say(ERROR, "Unexpected error (%s:%d).  Contact support and provide "
+			"debug log from %s.\n", file, line, DR_LOG_PATH);
+}
+
+void * __zalloc(size_t size, const char *func, int line)
+{
+	void *data;
+	data = malloc(size);
+	if (data)
+		memset(data, 0, size);
+	else
+		say(ERROR, "Allocation failure (%lx) at %s:%d\n", size, func, line);
+
+	return data;
+}
+
 /**
  * dr_init
  * @brief Initialization routine for drmgr and lsslot
  *
  */
-inline void
+inline int
 dr_init(void)
 {
+	int rc;
+
+	rc = dr_lock();
+	if (rc) {
+		say(ERROR, "Unable to obtain Dynamic Reconfiguration lock. "
+		    "Please try command again later.\n");
+		return -1;
+	}
+
+
 	log_fd = open(DR_LOG_PATH, O_RDWR | O_CREAT | O_APPEND,
 		      S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
 	if (log_fd == -1) {
-		fprintf(stderr, "Could not open log file %s\n\t%s\n",
-			DR_LOG_PATH, strerror(errno));
 		log_fd = 0;
+		say(ERROR, "Could not open log file %s\n\t%s\n", DR_LOG_PATH,
+		    strerror(errno));
 	} else {
 		time_t t;
 		char tbuf[128];
@@ -67,8 +122,16 @@ dr_init(void)
 		/* Insert seperator at beginning of drmgr invocation */
 		time(&t);
 		strftime(tbuf, 128, "%b %d %T %G", localtime(&t));
-		dbg("\n########## %s ##########\n", tbuf);
+		say(DEBUG, "\n########## %s ##########\n", tbuf);
 	}
+
+	/* Mask signals so we do not get interrupted */
+	if (sig_setup()) {
+		say(ERROR, "Could not mask signals to avoid interrupts\n");
+		return -1;
+	}
+
+	return 0;
 }
 
 /**
@@ -93,7 +156,7 @@ dr_fini(void)
 	/* Insert seperator at end of drmgr invocation */
 	time(&t);
 	strftime(tbuf, 128, "%b %d %T %G", localtime(&t));
-	dbg("########## %s ##########\n", tbuf);
+	say(DEBUG, "########## %s ##########\n", tbuf);
 
 	close(log_fd);
 
@@ -122,6 +185,42 @@ dr_fini(void)
 			return;
 		}
 	}
+
+	dr_unlock();
+}
+
+/**
+ * set_timeout
+ *
+ */
+void set_timeout(int timeout)
+{
+	if (!timeout) {
+		dr_timeout = -1;
+		return;
+	}
+
+	dr_timeout = time((time_t *)0);
+	if (dr_timeout == -1)
+		return;
+
+	dr_timeout += timeout;
+}
+
+/**
+ *
+ */
+int drmgr_timed_out(void)
+{
+	if (dr_timeout == -1)
+		return 0;	/* No timeout specified */
+
+	if (dr_timeout > time((time_t *)0))
+		return 0;
+
+	say(WARN, "Drmgr has exceeded its specified wait time and will not "
+	    "continue\n");
+	return 1;
 }
 
 /**
@@ -131,20 +230,15 @@ dr_fini(void)
  * This will attempt to lock a token (either file or directory) and wait
  * a specified amount of time if the lock cannot be granted.
  *
- * @param token token to lock
- * @param timeout timeout to wait before returning
  * @returns lock id if successful, -1 otherwise
  */
-int
-dr_lock(int timeout)
+int dr_lock(void)
 {
 	struct flock    dr_lock_info;
 	int             dr_lock_fd;
 	int             rc;
 	mode_t          old_mode;
-	long            start_time;	/* Beginning time. Used for timeout */
-	long            current_time;
-	int             first_try;
+	int             first_try = 1;
 
 	old_mode = umask(0);
 	dr_lock_fd = open(DR_LOCK_FILE, O_RDWR | O_CREAT,
@@ -158,24 +252,23 @@ dr_lock(int timeout)
 	dr_lock_info.l_start = 0;
 	dr_lock_info.l_len = 0;
 
-	first_try = 1;
-	start_time = time((long *) 0);	/* Pass 0 so that current time is
-					 * returned */
-	while (first_try || timeout == DR_WAIT ||
-	       timeout > (current_time = time((long *) 0)) - start_time)
-	{
-		if (!first_try)
+	do {
+		if (!first_try) {
 			sleep(1);
-		first_try = 0;
+			first_try = 0;
+		}
 
 		rc = fcntl(dr_lock_fd, F_SETLK, &dr_lock_info);
 		if (rc != -1)
 			return 0;
 
+		if (drmgr_timed_out())
+			break;
+
 		if (rc == -1 && errno == EACCES)
 			continue;
 
-	}
+	} while (1);
 
 	close(dr_lock_fd);
 	dr_lock_fd = 0;
@@ -237,11 +330,12 @@ add_node(char *path, struct of_node *new_nodes)
 	 * device tree.
 	 */
 	if (!stat(path, &sbuf)) {
-		dbg("Device-tree node %s already exists, skipping\n", path);
+		say(DEBUG, "Device-tree node %s already exists, skipping\n",
+		    path);
 		return 0;
 	}
 
-	dbg("Adding device-tree node %s\n", path);
+	say(DEBUG, "Adding device-tree node %s\n", path);
 
 	/* The path passed in is a full ofdt path, remove the preceeding
 	 * /proc/device-tree for the write to the kernel.
@@ -263,13 +357,13 @@ add_node(char *path, struct of_node *new_nodes)
 	}
 
 	if (! nprops) {
-		err_msg("new_nodes have no properties\n");
+		say(ERROR, "new_nodes have no properties\n");
 		return -1;
 	}
 
 	buf = zalloc(bufsize);
 	if (buf == NULL) {
-		err_msg("Failed to allocate buffer to write to kernel\n");
+		say(ERROR, "Failed to allocate buffer to write to kernel\n");
 		return errno;
 	}
 
@@ -300,17 +394,19 @@ add_node(char *path, struct of_node *new_nodes)
 	*pos = '\0';
 
 	/* dump the buffer for debugging */
-	dbg("ofdt update: %s\n", buf);
+	say(DEBUG, "ofdt update: %s\n", buf);
 
 	fd = open(OFDTPATH, O_WRONLY);
 	if (fd <= 0) {
-		err_msg("Failed to open %s: %s\n", OFDTPATH, strerror(errno));
+		say(ERROR, "Failed to open %s: %s\n", OFDTPATH,
+		    strerror(errno));
 		return errno;
 	}
 
 	rc = write(fd, buf, bufsize);
 	if (rc <= 0)
-		err_msg("Write to %s failed: %s\n", OFDTPATH, strerror(errno));
+		say(ERROR, "Write to %s failed: %s\n", OFDTPATH,
+		    strerror(errno));
 	else
 		rc = 0;
 
@@ -336,7 +432,7 @@ remove_node(char *path)
 	int cmdlen;
 	char buf[DR_PATH_MAX];
 
-	dbg("Removing device-tree node %s\n", path);
+	say(DEBUG, "Removing device-tree node %s\n", path);
 
 	memset(buf, 0, DR_PATH_MAX);
 
@@ -349,13 +445,15 @@ remove_node(char *path)
 
 	fd = open(OFDTPATH, O_WRONLY);
 	if (fd <= 0) {
-		err_msg("Failed to open %s: %s\n", OFDTPATH, strerror(errno));
+		say(ERROR, "Failed to open %s: %s\n", OFDTPATH,
+		    strerror(errno));
 		return errno;
 	}
 
 	rc = write(fd, buf, cmdlen);
 	if (rc != cmdlen)
-		err_msg("Write to %s failed: %s\n", OFDTPATH, strerror(errno));
+		say(ERROR, "Write to %s failed: %s\n", OFDTPATH,
+		    strerror(errno));
 	else
 		rc = 0;
 
@@ -444,7 +542,7 @@ remove_device_tree_nodes(char *path)
 
 	d = opendir(path);
 	if (d == NULL) {
-		err_msg("Could not open %s: %s\n", path, strerror(errno));
+		say(ERROR, "Could not open %s: %s\n", path, strerror(errno));
 		return -1;
 	}
 
@@ -494,17 +592,19 @@ update_property(const char *buf, size_t len)
 {
 	int fd, rc;
 
-	dbg("Updating of property\n");
+	say(DEBUG, "Updating OF property\n");
 
 	fd = open(OFDTPATH, O_WRONLY);
 	if (fd <= 0) {
-		err_msg("Failed to open %s: %s\n", OFDTPATH, strerror(errno));
+		say(ERROR, "Failed to open %s: %s\n", OFDTPATH,
+		    strerror(errno));
 		return errno;
 	}
 
 	rc = write(fd, buf, len);
 	if ((size_t)rc != len)
-		err_msg("Write to %s failed: %s\n", OFDTPATH, strerror(errno));
+		say(ERROR, "Write to %s failed: %s\n", OFDTPATH,
+		    strerror(errno));
 	else
 		rc = 0;
 
@@ -611,6 +711,27 @@ get_str_attribute(const char *path, const char *attribute, void *buf,
 	return get_att_prop(path, attribute, buf, buf_sz, "%s");
 }
 
+/**
+ * get_ofdt_uint_property
+ * @brief retrieve an unsigned integer property from the device tree and
+ * byteswap if needed (device tree is big endian).
+ *
+ * @param path path to the property to retrieve
+ * @param name name of the property to retrieve
+ * @param data unsigned integer pointer to write property to
+ * @returns 0 on success, -1 otherwise
+ */
+int
+get_ofdt_uint_property(const char *path, const char *attribute, uint *data)
+{
+	uint tmp;
+	int rc;
+	rc = get_property(path, attribute, &tmp, sizeof(tmp));
+	if (!rc)
+		*data = be32toh(tmp);
+	return rc;
+}
+
 /**
  * get_property_size
  * @brief retrieve the size of a property
@@ -643,7 +764,8 @@ get_property_size(const char *path, const char *property)
 void
 sighandler(int signo)
 {
-	err_msg("Received signal %d, attempting to cleanup and exit\n", signo);
+	say(ERROR, "Received signal %d, attempting to cleanup and exit\n",
+	    signo);
 
 	if (log_fd) {
 		void *callstack[128];
@@ -653,7 +775,6 @@ sighandler(int signo)
 		backtrace_symbols_fd(callstack, sz, log_fd);
 	}
 
-	dr_unlock();
 	dr_fini();
 	exit(-1);
 }
@@ -755,17 +876,29 @@ char *php_slot_type_msg[]={
 	"PCI-E capable, Rev 2, 8x lanes",
 	"PCI-E capable, Rev 2, 16x lanes",
 	"PCI-E capable, Rev 2, 32x lanes",	/* 30 */
+	"PCI-E capable, Rev 3, 8x lanes with 1 lane connected",
+	"PCI-E capable, Rev 3, 8x lanes with 4x lanes connected",
+	"PCI-E capable, Rev 3, 8x lanes with 8x lanes connected",
+	"PCI-E capable, Rev 3, 16x lanes with 1 lane connected",
+	"PCI-E capable, Rev 3, 16x lanes with 8x lanes connected",  /* 35 */
+	"PCI-E capable, Rev 3, 16x lanes with 16x lanes connected",
+	"PCI-E capable, Rev 4, 8x lanes with 1 lane connected",
+	"PCI-E capable, Rev 4, 8x lanes with 4x lanes connected",
+	"PCI-E capable, Rev 4, 8x lanes with 8x lanes connected",
+	"PCI-E capable, Rev 4, 16x lanes with 1 lane connected",    /* 40 */
+	"PCI-E capable, Rev 4, 16x lanes with 8x lanes connected",
+	"PCI-E capable, Rev 4, 16x lanes with 16x lanes connected",
 };
 
 char *
 node_type(struct dr_node *node)
 {
 	int desc_msg_num;
-	char *desc;
+	char *desc = "Unknown";
 
 	desc_msg_num = atoi(node->drc_type);
 	if ((desc_msg_num >= 1 &&  desc_msg_num <= 8) ||
-	    (desc_msg_num >= 11 && desc_msg_num <= 30))
+	    (desc_msg_num >= 11 && desc_msg_num <= 42))
 		desc = php_slot_type_msg[desc_msg_num];
 	else {
 		switch (node->dev_type) {
@@ -805,13 +938,14 @@ valid_platform(const char *platform)
 
 	rc = get_property(OFDT_BASE, "device_type", buf, 128);
 	if (rc) {
-		err_msg("Cannot open %s: %s\n", PLATFORMPATH, strerror(errno));
+		say(ERROR, "Cannot open %s: %s\n", PLATFORMPATH,
+		    strerror(errno));
 		return 0;
 	}
 
 	if (strcmp(buf, platform)) {
-		err_msg("This command is not supported for %s platforms.\n",
-			platform);
+		say(ERROR, "This command is not supported for %s platforms.\n",
+		    platform);
 		return 0;
 	}
 
@@ -834,7 +968,7 @@ get_sysparm(const char *parm, unsigned long *value)
 
 	f = fopen(LPARCFG_PATH, "r");
 	if (f == NULL) {
-		dbg("Could not open \"%s\"\n%s\n", LPARCFG_PATH,
+		say(ERROR, "Could not open \"%s\"\n%s\n", LPARCFG_PATH,
 		    strerror(errno));
 		return -1;
 	}
@@ -854,7 +988,7 @@ get_sysparm(const char *parm, unsigned long *value)
 
 	fclose(f);
 	if (rc)
-		dbg("Error finding %s in %s\n", parm, LPARCFG_PATH);
+		say(ERROR, "Error finding %s in %s\n", parm, LPARCFG_PATH);
 
 	return rc;
 }
@@ -874,17 +1008,17 @@ set_sysparm(const char *parm, int val)
 
 	f = fopen(LPARCFG_PATH, "w");
 	if (f == NULL) {
-		dbg("Could not open \"%s\"\n%s\n", LPARCFG_PATH,
+		say(ERROR, "Could not open \"%s\"\n%s\n", LPARCFG_PATH,
 		    strerror(errno));
 		return -1;
 	}
 
-	dbg("Updating sysparm %s to %d...", parm, val);
+	say(DEBUG, "Updating sysparm %s to %d...", parm, val);
 	rc = fprintf(f, "%s=%d\n", parm, val);
 
 	fclose(f);
 
-	dbg("%s.\n", (rc == -1) ? "fail" : "success");
+	say(DEBUG, "%s.\n", (rc == -1) ? "fail" : "success");
 	return (rc == -1) ? -1 : 0;
 }
 
@@ -945,23 +1079,23 @@ update_sysparm(struct options *opts)
 	/* Validate capability */
 	if (! strcmp(opts->ctype, "cpu")) {
 		if (! cpu_entitlement_capable()) {
-			err_msg("CPU entitlement capability is not enabled "
-				"on this platform.\n");
+			say(ERROR, "CPU entitlement capability is not enabled "
+			    "on this platform.\n");
 			return -1;
 		}
 
 		sysparm_table = cpu_sysparm_table;
 	} else if (! strcmp(opts->ctype, "mem")) {
 		if (! mem_entitlement_capable()) {
-			err_msg("Memory entitlement capability is not enabled "
-				"on this platform.\n");
+			say(ERROR, "Memory entitlement capability is not "
+			    "enabled on this platform.\n");
 			return -1;
 		}
 
 		sysparm_table = mem_sysparm_table;
 	} else {
-		err_msg("Invalid entitlement update type \"%s\" specified.\n",
-			opts->ctype);
+		say(ERROR, "Invalid entitlement update type \"%s\" "
+		    "specified.\n", opts->ctype);
 		return -1;
 	}
 	
@@ -979,23 +1113,23 @@ update_sysparm(struct options *opts)
 	}
 
 	if (linux_parm == NULL) {
-		err_msg("The entitlement parameter \"%s\" is not recognized\n",
-			opts->p_option);
+		say(ERROR, "The entitlement parameter \"%s\" is not "
+		    "recognized\n", opts->p_option);
 		return -1;
 	}
 
 	if ((get_sysparm(linux_parm, &curval)) < 0) {
-		err_msg("Could not get current system parameter value of "
-			"%s (%s)\n", linux_parm, opts->p_option);
+		say(ERROR, "Could not get current system parameter value of "
+		    "%s (%s)\n", linux_parm, opts->p_option);
 		return -1;
 	}
 
 	if (opts->action == REMOVE) {
 		if (opts->quantity > curval) {
-			err_msg("Cannot reduce system parameter value %s by "
-				"more than is currently available.\nCurrent "
-				"value: %lx, asking to remove: %x\n",
-				opts->p_option, curval, opts->quantity);
+			say(ERROR, "Cannot reduce system parameter value %s by "
+			    "more than is currently available.\nCurrent "
+			    "value: %lx, asking to remove: %x\n",
+			    opts->p_option, curval, opts->quantity);
 			return 1;
 		}
 
@@ -1015,11 +1149,12 @@ cpu_dlpar_capable(void)
 	char *cpu_dir = "/sys/devices/system/cpu";
 	int capable = 1;
 
-	dbg("Validating CPU DLPAR capability...");
+	say(ERROR, "Validating CPU DLPAR capability...");
 	
 	d = opendir(cpu_dir);
 	if (d == NULL) {
-		dbg("no.\n    opendir(\"%s\"): %s\n", cpu_dir, strerror(errno));
+		say(ERROR, "no.\n    opendir(\"%s\"): %s\n", cpu_dir,
+		    strerror(errno));
 		return 0;
 	}
 
@@ -1036,12 +1171,12 @@ cpu_dlpar_capable(void)
 		sprintf(fname, "%s/%s/online", cpu_dir, de->d_name);
 		
 		if (stat(fname, &sbuf)) {
-			dbg("no.\n    stat(\"%s\"): %s\n", fname,
+			say(ERROR, "no.\n    stat(\"%s\"): %s\n", fname,
 			    strerror(errno));
 			capable = 0;
 		}
 
-		dbg("yes.\n");
+		say(ERROR, "yes.\n");
 		break;
 	}
 
@@ -1055,13 +1190,14 @@ dlpar_capable(const char *type, const char *fname)
 	struct stat sbuf;
 	int capable = 1;
 
-	dbg("Validating %s capability...", type);
+	say(ERROR, "Validating %s capability...", type);
 	
 	if (stat(fname, &sbuf)) {
-		dbg("no.\n    stat(\"%s\"): %s\n", fname, strerror(errno));
+		say(ERROR, "no.\n    stat(\"%s\"): %s\n", fname,
+		    strerror(errno));
 		capable = 0;
 	} else {
-		dbg("yes.\n");
+		say(ERROR, "yes.\n");
 	}
 
 	return capable;
@@ -1077,8 +1213,19 @@ mem_dlpar_capable(void)
 static int
 check_slot_phb_dlpar(const char *type)
 {
+	struct stat sbuf;
 	int rc;
 
+	/* Before checking for dlpar capability, we need to ensure that
+	 * rpadlpar_io module is loaded or built into the kernel. This
+	 * does make the checking a bit redundant though.
+	 */
+	if ((stat(add_slot_fname, &sbuf)) || (stat(ADD_SLOT_FNAME2, &sbuf))) {
+		rc = system("/sbin/modprobe rpadlpar_io");
+		if (WIFEXITED(rc) && WEXITSTATUS(rc))
+			say(ERROR, "rpadlpar_io module was not loaded\n");
+	}
+
 	/* For unknown reasons the add_slot and remove_slot sysfs files
 	 * used for dlpar operations started appearing with quotes around
 	 * the filename.  So, this little hack exists to ensure nothing
@@ -1087,12 +1234,12 @@ check_slot_phb_dlpar(const char *type)
 	 * The default value is without the quotes.  This is what was and
 	 * what shall be again.
 	 */
-	rc = dlpar_capable(type, dlpar_add_slot);
+	rc = dlpar_capable(type, add_slot_fname);
 	if (! rc) {
-		dlpar_add_slot = DLPAR_ADD_SLOT2;
-		dlpar_remove_slot = DLPAR_REMOVE_SLOT2;
+		add_slot_fname = ADD_SLOT_FNAME2;
+		remove_slot_fname = REMOVE_SLOT_FNAME2;
 		
-		rc = dlpar_capable(type, dlpar_add_slot);
+		rc = dlpar_capable(type, add_slot_fname);
 	}
 	
 	return rc;
@@ -1210,13 +1357,18 @@ int ams_balloon_active(void)
 	char *ams_enabled = "/sys/devices/system/cmm/cmm0/loaned_kb";
 	char *cmm_param_path = "/sys/module/cmm/parameters";
 	struct stat sbuf;
-	int is_inactive = 1;
+	static int is_inactive = 1;
+	static int ams_checked = 0;
 
-	if (!stat(ams_enabled, &sbuf) && !stat(cmm_param_path, &sbuf)) {
-		get_int_attribute(cmm_param_path, "disable",
-				  &is_inactive, sizeof(is_inactive));
+	if (!ams_checked) {
+		if (!stat(ams_enabled, &sbuf) && !stat(cmm_param_path, &sbuf))
+			get_int_attribute(cmm_param_path, "disable",
+					  &is_inactive, sizeof(is_inactive));
+
+		say(DEBUG, "AMS ballooning %s active\n",
+		    is_inactive?"is not":"is");
+		ams_checked = 1;
 	}
 
-	dbg("AMS ballooning %s active\n", is_inactive?"is not":"is");
 	return !is_inactive;
 }
