@@ -205,7 +205,7 @@ get_mem_node_lmbs(struct lmb_list_head *lmb_list)
 
 		if (lmb == NULL) {
 			say(DEBUG, "Could not find LMB with drc-index of %x\n",
-			    lmb->drc_index);
+			    my_drc_index);
 			rc = -1;
 			break;
 		}
@@ -298,28 +298,23 @@ get_dynamic_reconfig_lmbs(struct lmb_list_head *lmb_list)
 	for (i = 0; i < num_entries; i++) {
 		struct dr_node *lmb;
 
-		/* convert for LE systems */
-		drmem->address = be64toh(drmem->address);
-		drmem->drc_index = be32toh(drmem->drc_index);
-		drmem->flags = be32toh(drmem->flags);
-
 		for (lmb = lmb_list->lmbs; lmb; lmb = lmb->next) {
-			if (lmb->drc_index == drmem->drc_index)
+			if (lmb->drc_index == be32toh(drmem->drc_index))
 				break;
 		}
 
 		if (lmb == NULL) {
 			say(DEBUG, "Could not find LMB with drc-index of %x\n",
-			    lmb->drc_index);
+			    drmem->drc_index);
 			rc = -1;
 			break;
 		}
 
 		sprintf(lmb->ofdt_path, DYNAMIC_RECONFIG_MEM);
 		lmb->lmb_size = lmb_sz;
-		lmb->lmb_address = drmem->address;
+		lmb->lmb_address = be64toh(drmem->address);
 
-		if (drmem->flags & DRMEM_ASSIGNED) {
+		if (be32toh(drmem->flags) & DRMEM_ASSIGNED) {
 			found++;
 			lmb->is_owned = 1;
 
@@ -576,19 +571,15 @@ update_drconf_node(struct dr_node *lmb, struct lmb_list_head *lmb_list,
 
 	for (i = 0; i < entries; i++) {
 
-		/* convert for LE systems */
-		drmem->drc_index = be32toh(drmem->drc_index);
-		drmem->flags = be32toh(drmem->flags);
-
-		if (drmem->drc_index != lmb->drc_index) {
+		if (be32toh(drmem->drc_index) != lmb->drc_index) {
 			drmem++;
 			continue;
 		}
 
 		if (action == ADD)
-			drmem->flags |= DRMEM_ASSIGNED;
+			drmem->flags |= be32toh(DRMEM_ASSIGNED);
 		else
-			drmem->flags &= ~DRMEM_ASSIGNED;
+			drmem->flags &= be32toh(~DRMEM_ASSIGNED);
 
 		break;
 	}
@@ -1211,12 +1202,21 @@ int
 valid_mem_options(struct options *opts)
 {
 	/* default to a quantity of 1 */
-	if ((opts->quantity == 0) && (opts->usr_drc_name == NULL))
+	if (opts->quantity == 0)
 		opts->quantity = 1;
 
-	if ((opts->action != ADD) && (opts->action != REMOVE))
+	if ((opts->action != ADD) && (opts->action != REMOVE)) {
 		say(ERROR, "The '-r' or '-a' option must be specified for "
 		    "memory operations\n");
+		return -1;
+	}
+
+	/* The -s option can specify a drc name or drc index */
+	if (opts->usr_drc_name && !strncmp(opts->usr_drc_name, "0x", 2)) {
+		opts->usr_drc_index = strtoul(opts->usr_drc_name, NULL, 16);
+		opts->usr_drc_name = NULL;
+	}
+
 	return 0;
 }
 
